import setuptools
from distutils.core import setup

setup(
    name='hdtools',
    version='0.1.1',
    description='hd-tools for bitcoin and bitcoin-test',
    author='Mahdi Fooladgar',
    author_email='fooladgar@morvarid.io',
    url='https://github.com/morvaridio/hdtools',
    license='MIT',
    packages=[
        'hdtools',
    ]
)
